import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Steamship


@pytest.mark.usefixtures("client")
def test_web(client: Steamship):

    web_agent_path = PACKAGES_PATH / "transports" / "test_web_agent.py"

    with deploy_package(
        client,
        web_agent_path,
    ) as (_, _, web_instance):

        response = web_instance.invoke("answer", question="test")
        assert response[0].get("text") == "Response to: test"

        # Test multiple times since we previously had issues on subsequent calls
        response = web_instance.invoke("answer", question="test2")
        assert response[0].get("text") == "Response to: test2"

        response = web_instance.invoke("answer", question="test3")
        assert response[0].get("text") == "Response to: test3"



from __future__ import annotations

import logging
from functools import partial
from typing import Any, Dict, List, Optional, Type

from steamship import SteamshipError, Task
from steamship.base.package_spec import MethodSpec, PackageSpec, RouteConflictError
from steamship.invocable import Invocable

# Note!
# =====
#
# This the files in this package are for Package Implementors.
# If you are using the Steamship Client, you probably are looking for either steamship.client or steamship.data
#
from steamship.invocable.invocable import find_route_methods
from steamship.invocable.package_mixin import PackageMixin
from steamship.utils.url import Verb


class PackageService(Invocable):
    """The Abstract Base Class of a Steamship Package.

    Packages may implement whatever methods they like.  To expose these methods as invocable HTTP routes,
    annotate the method with @get or @post and the route name.

    Package *implementations* are effectively stateless, though they will have stateful

    """

    USED_MIXIN_CLASSES: List[Type[PackageMixin]] = []
    mixins: List[PackageMixin]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mixins = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # The subclass takes care to extend what the superclass may have set here by copying it.
        _package_spec = PackageSpec(name=cls.__name__, doc=cls.__doc__, methods=[])
        if hasattr(cls, "_package_spec"):
            _package_spec.import_parent_methods(cls._package_spec)

        # The subclass always overrides whatever the superclass set here, having cloned its data.
        cls._package_spec = _package_spec

        # Now must add routes for mixins
        for used_mixin_class in cls.USED_MIXIN_CLASSES:
            cls.scan_mixin(cls._package_spec, used_mixin_class)

    @classmethod
    def scan_mixin(
        cls,
        package_spec: PackageSpec,
        mixin_class: Type[PackageMixin],
        mixin_instance: Optional[PackageMixin] = None,
        permit_overwrite_of_existing_methods: bool = False,
    ):
        for route_method in find_route_methods(mixin_class):
            if mixin_instance is not None:
                func_binding = partial(route_method.attribute, self=mixin_instance)
            else:
                func_binding = (
                    route_method.attribute
                )  # This binding is not truly valid. It must be overwritten during add_mixin
            method_spec = MethodSpec.from_class(
                mixin_class,
                route_method.attribute.__name__,
                path=route_method.path,
                verb=route_method.verb,
                config=route_method.config,
                func_binding=func_binding,
            )
            try:
                allow_override_for_this_route = permit_overwrite_of_existing_methods
                existing_route = package_spec.method_mappings.get(route_method.verb, {}).get(
                    "/" + route_method.path
                )
                if (
                    existing_route is not None
                    and existing_route.class_name == mixin_class.__name__
                    and mixin_instance is not None
                ):
                    allow_override_for_this_route = True
                package_spec.add_method(
                    method_spec,
                    permit_overwrite_of_existing=allow_override_for_this_route,
                )
            except RouteConflictError as conflict_error:
                message = f"When attempting to add mixin {mixin_class.__name__}, route {route_method.verb} {route_method.path} conflicted with already added route {route_method.verb} {route_method.path} on class {conflict_error.existing_method_spec.class_name}"
                raise RouteConflictError(
                    message=message,
                    existing_method_spec=conflict_error.existing_method_spec,
                )

    def add_mixin(self, mixin: PackageMixin, permit_overwrite_of_existing_methods: bool = False):
        PackageService.scan_mixin(
            self._package_spec,
            mixin.__class__,
            mixin,
            permit_overwrite_of_existing_methods=permit_overwrite_of_existing_methods,
        )
        self.mixins.append(mixin)

    def instance_init(self):
        """The instance init method will be called ONCE by the engine when a new instance of a package or plugin has been created. By default, this calls instance_init on mixins, in order."""
        for mixin in self.mixins:
            mixin.instance_init(self.config, self.context)

    def invoke_later(
        self,
        method: str,
        verb: Verb = Verb.POST,
        wait_on_tasks: List[Task] = None,
        arguments: Dict[str, Any] = None,
        delay_ms: Optional[int] = None,
    ) -> Task[Any]:
        """Schedule a method for future invocation.

        Parameters
        ----------
        method: str
                The method to invoke, as registered with Steamship in the @get or @post decorator.
        verb:   Verb
                The HTTP Verb to use. Default is POST.
        wait_on_tasks: List[Task]
                A list of Task objects (or task IDs) that should be waited upon before invocation.
        arguments: Dict[str, Any]
                The keyword arguments of the invoked method
        delay_ms: Optional[int]
                A delay, in milliseconds, before the invocation should execute.

        Returns
        -------
        Task[Any]
                A Task representing the future work
        """

        if self.context is None:
            raise SteamshipError(
                message="Unable to call invoke_later because the InvocationContext was None"
            )
        if self.context.invocable_instance_handle is None:
            raise SteamshipError(
                message="Unable to call invoke_later because the invocable_instance_handle on InvocationContext was None"
            )

        payload = {
            "instanceHandle": self.context.invocable_instance_handle,
            "payload": {
                "httpVerb": verb.value,
                "invocationPath": method,
                "arguments": arguments or {},
            },
        }
        operation = "package/instance/invoke"

        logging.info(
            f"Scheduling {verb} {method} for future invocation on me ({self.context.invocable_handle})"
        )

        resp = self.client.post(
            operation,
            payload,
            expect=Task[Task],  # This operation should return a task
            as_background_task=True,  # This operation should always be asynchronous
            wait_on_tasks=wait_on_tasks,  # This operation might await other tasks first
            task_delay_ms=delay_ms,  # This operation might have a required delay
        )
        return resp



from steamship import File
from steamship.invocable import PackageService, post
from steamship.invocable.mixins.blockifier_mixin import BlockifierMixin
from steamship.invocable.mixins.file_importer_mixin import FileImporterMixin
from steamship.invocable.mixins.indexer_mixin import IndexerMixin


class PackageWithImporterMixin(PackageService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_mixin(FileImporterMixin(self.client))
        self.add_mixin(BlockifierMixin(self.client))
        self.add_mixin(IndexerMixin(self.client))

    @post("get_file")
    def get_file(self, id: str) -> File:
        file = File.get(self.client, _id=id)
        return file



from typing import List, Optional, Tuple

import requests

from steamship import DocTag, File, MimeTypes, Steamship, SteamshipError, Tag, Task
from steamship.agents import logging
from steamship.invocable import post
from steamship.invocable.package_mixin import PackageMixin
from steamship.utils.file_tags import update_file_status


class FileImporterMixin(PackageMixin):
    """Provide endpoints for easy file import -- both sync and async."""

    client: Steamship

    def __init__(self, client: Steamship):
        self.client = client

    def _async_importer_for_url(self, url: str) -> Optional[str]:
        """Return the async importer plugin, if necessary."""
        if "youtube.com" in url or "youtu.be" in url:
            return "youtube-file-importer"
        return None

    def _import_with_async_importer(
        self,
        url: str,
        importer_handle: str,
        mime_type: Optional[str] = None,
        tags: Optional[List[Tag]] = None,
    ) -> Tuple[File, Optional[Task]]:
        """Import a URL via an async FileImporter, returning a synchronous File but async import Task."""

        if ("youtube" in url or "youtu.be" in url) and mime_type is None:
            # Mark it as audio so that the s2t will work later.
            mime_type = MimeTypes.MP3

        file = File.create(self.client, tags=tags, mime_type=mime_type)

        if importer_handle is None:
            raise SteamshipError(
                message=f"Unable to async auto-guess file importer for {url} and none was provided."
            )

        file_importer = self.client.use_plugin(importer_handle)
        task = file.import_with_plugin(
            plugin_instance=file_importer.handle, url=url, mime_type=mime_type
        )
        update_file_status(self.client, file, "Importing")

        return file, task

    def _scrape_and_import_url(
        self,
        url: str,
        mime_type: Optional[str] = None,
        tags: Optional[List[Tag]] = None,
    ) -> Tuple[File, Optional[Task]]:
        """Scrape and then import the URL to a File, returning a synchronous File and no import Task."""
        if mime_type is None and ".pdf" in url:
            mime_type = MimeTypes.PDF

        response = requests.get(url)
        _bytes = response.content

        if not response.ok:
            msg = f"Error importing url {url}. Response was {response.text}"
            logging.error(msg)
            raise SteamshipError(message=msg)

        file = File.create(self.client, content=_bytes, mime_type=mime_type, tags=tags)
        return file, None

    def import_url_to_file_and_task(self, url: str) -> Tuple[File, Optional[Task]]:
        """Import the provided URL, returning the file and optional task, if async work is required."""
        async_importer_for_url = self._async_importer_for_url(url)

        source_tag = Tag(kind=DocTag.SOURCE, name=url)

        # Hacky way to get ta title
        title = url.split("/")[-1]
        title = title.split("?")[0]
        title = title.split("#")[0]

        title_tag = Tag(kind=DocTag.TITLE, name=title)
        tags = [source_tag, title_tag]

        if async_importer_for_url:
            return self._import_with_async_importer(
                url, async_importer_for_url, tags=tags, mime_type=None
            )
        else:
            return self._scrape_and_import_url(url, tags=tags)

    @post("/import_url")
    def import_url(self, url: str) -> File:
        """Import the URL to a Steamship File. Actual import will be scheduled Async."""
        file, task = self.import_url_to_file_and_task(url)
        return file

    @post("/import_text")
    def import_text(self, text: str, mime_type: Optional[str]) -> File:
        """Import the text to a Steamship File."""
        return File.create(self.client, content=text, mime_type=mime_type)



from abc import ABC, abstractmethod
from typing import List

from steamship import Tag


class TextSplitter(ABC):
    @abstractmethod
    def split_text(self, text: str) -> List[str]:
        """Split the incoming text into strings"""
        raise NotImplementedError()

    def chunk_text_to_tags(self, text: str, kind: str, name: str = None) -> List[Tag]:
        """Split the incoming text into strings, and then wrap those strings in Tags"""
        if text is not None and text != "":
            text_splits = self.split_text(text)
            start_index = 0
            result = []
            for text_split in text_splits:
                result.append(
                    Tag(
                        kind=kind,
                        name=name,
                        start_idx=start_index,
                        end_idx=start_index + len(text_split),
                        text=text_split,
                    )
                )
                start_index += len(text_split)
            return result
        else:
            return []


class FixedSizeTextSplitter(TextSplitter):
    """Simplest possible chunking strategy; every n characters."""

    chunk_size: int

    def __init__(self, chunk_size):
        self.chunk_size = chunk_size

    def split_text(self, text: str) -> List[str]:
        result = []
        for i in range(int(len(text) / self.chunk_size) + 1):
            start = i * self.chunk_size
            end = min((i + 1) * self.chunk_size, len(text))
            result.append(text[start:end])
        return result



from typing import Optional

from steamship import File, MimeTypes, Steamship, SteamshipError, Task
from steamship.invocable import post
from steamship.invocable.package_mixin import PackageMixin
from steamship.utils.file_tags import update_file_status


class BlockifierMixin(PackageMixin):
    """Provides endpoints for easy Blockification of files."""

    client: Steamship

    def __init__(self, client: Steamship):
        self.client = client

    @post("/blockify_file")
    def blockify(
        self,
        file_id: str,
        mime_type: Optional[MimeTypes] = None,
        blockifier_handle: Optional[str] = None,
        after_task_id: Optional[str] = None,
    ) -> Task:
        """Blockify the file `file_id` using a curated set of defaults for the provided `mime_type`.

        If no MIME Type is provided, the file's recorded MIME Type will be used.
        If still no MIME Type is available, an error will be thrown.

        Supported MIME Types:

        - PDF
        - Audio (MP3, MP4, WEBM)
        """

        file = File.get(self.client, _id=file_id)
        update_file_status(self.client, file, "Blockifying")

        _mime_type = mime_type or file.mime_type
        if not _mime_type:
            raise SteamshipError(
                message=f"No MIME Type found for file {file.id}. Unable to blockify."
            )

        plugin_instance = None

        if blockifier_handle:
            plugin_instance = self.client.use_plugin(blockifier_handle)
        elif _mime_type == MimeTypes.PDF:
            plugin_instance = self.client.use_plugin("pdf-blockifier")
        elif _mime_type in [MimeTypes.MP3, MimeTypes.MP4_AUDIO, MimeTypes.WEBM_AUDIO]:
            plugin_instance = self.client.use_plugin("s2t-blockifier-default")
        elif _mime_type in [MimeTypes.MKD, MimeTypes.TXT]:
            plugin_instance = self.client.use_plugin("markdown-blockifier-default")

        if not plugin_instance:
            raise SteamshipError(
                message=f"Unable to blockify file {file.id}. MIME Type {_mime_type} unsupported"
            )

        # Postpone the operation if required.
        wait_on_tasks = []
        if after_task_id:
            wait_on_tasks.append(Task(client=self.client, task_id=after_task_id))

        return file.blockify(plugin_instance.handle, wait_on_tasks=wait_on_tasks)



import pytest
from steamship_tests import PACKAGES_PATH, TEST_ASSETS_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import File, MimeTypes, Steamship, Task, TaskState
from steamship.data.plugin.index_plugin_instance import SearchResults


@pytest.mark.usefixtures("client")
def test_importer_mixin_and_package_invocation(client: Steamship):
    demo_package_path = PACKAGES_PATH / "package_with_mixin_importer.py"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        mixin_response = instance.invoke("import_url", url="https://www.google.com/")
        assert mixin_response
        file_id = mixin_response.get("id")
        assert file_id
        file = File.get(client, _id=file_id)
        content = file.raw()
        assert content

        # Now import content
        text_content = """# Title

Hi there this is a paragraph.
        """

        mixin_response2 = instance.invoke(
            "import_text", text=text_content, mime_type="text/markdown"
        )
        assert mixin_response2
        file_id = mixin_response2.get("id")
        assert file_id
        file2 = File.get(client, _id=file_id)
        content2 = file2.raw().decode("utf-8")
        assert content2 == text_content
        assert file2.mime_type == "text/markdown"

        # Test the blockify mixin

        blockify_task = instance.invoke("blockify_file", file_id=file_id)
        assert blockify_task

        blockify_task = Task.parse_obj(blockify_task)
        blockify_task.client = client
        assert blockify_task.task_id
        assert blockify_task.state == TaskState.waiting
        blockify_task.wait()

        file3 = file2.refresh()
        assert file3.blocks
        assert len(file3.blocks) == 2

        # Test blockifier with a PDF

        pdf_path = TEST_ASSETS_PATH / "test.pdf"

        with pdf_path.open("rb") as f:
            pdf_bytes = f.read()

        pdf_file = File.create(client, content=pdf_bytes, mime_type=MimeTypes.PDF)
        assert pdf_file.mime_type == MimeTypes.PDF
        assert pdf_file.id is not None
        content = pdf_file.raw()
        assert content is not None

        pdf_blockify_task = instance.invoke("blockify_file", file_id=pdf_file.id)
        assert pdf_blockify_task

        pdf_blockify_task = Task.parse_obj(pdf_blockify_task)
        pdf_blockify_task.client = client
        assert pdf_blockify_task.task_id
        assert pdf_blockify_task.state == TaskState.waiting
        pdf_blockify_task.wait()

        pdf_file = pdf_file.refresh()
        assert len(pdf_file.blocks) > 0

        pdf_blocks = pdf_file.blocks
        assert pdf_blocks
        assert len(pdf_blocks) == 2
        assert pdf_blocks[0].text == "This is the ﬁrst page\n"
        assert pdf_blocks[1].text == "This is the second page"

        # Test indexer with text
        instance.invoke("index_text", text="Bananas")
        instance.invoke("index_text", text="Trains")
        instance.invoke("index_text", text="Democracy")

        instance.invoke("index_text", text="Pizza", index_handle="index2")
        instance.invoke("index_text", text="Mountains", index_handle="index2")

        result = instance.invoke("search_index", query="politics")
        result = SearchResults.parse_obj(result)
        winner = result.items[0]
        assert winner.tag.text in ["Trains", "Democracy", "Bananas"]

        result = instance.invoke("search_index", query="passengers", index_handle="index2")
        result = SearchResults.parse_obj(result)
        winner = result.items[0]
        assert winner.tag.text in ["Pizza", "Mountains"]

        # Test indexer with file
        instance.invoke("index_file", file_id=pdf_file.id, index_handle="i2")

        result = instance.invoke("search_index", query="second page", index_handle="i2")
        result = SearchResults.parse_obj(result)
        assert len(result.items) == 2
        winner = result.items[0]
        assert winner.tag.text == "This is the second page"

        other = result.items[1]
        assert other.tag.text == "This is the ﬁrst page\n"



import pytest
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Package, Steamship


@pytest.mark.usefixtures("client")
def test_package_create(client: Steamship):
    client = get_steamship_client()
    _ = Package.create(client)
